/* litmus_so.h - Main include file of the LITMUS Shared Object API. 
 *
 * This the main API to LITMUS shared objects. LITMUS shared objects
 * are binary files of fixed size that serve as a shared memory
 * pool. They are intended to be mapped into the address space of
 * cooperating processes via the mmap(2) system call.  
 * 
 * The files consist of a management header and a heap space. In
 * the management header information on the current state of the
 * object is kept. For example it includes  memory management
 * information and information on how many processes have mapped the
 * file. Applications should not attempt to manipulate the
 * header directly. Instead, they should use the API provided in this file.
 */

#ifndef __LITMUS_SO_H__
#define __LITMUS_SO_H__

/* lso_t - A handle to a litmus shared object. 
 *         Shared objects are opaque.
 */
typedef struct litmus_so*  lso_t;

/* lso_global_t - An address that is meaningful for all processes
 *                sharing the object.
 */
typedef unsigned int lso_global_t;

typedef enum {
	/* no error occured */
	LSO_OK = 0,

	/* shared object does not exist */
	LSO_NO_FILE,

	/* the file pointed to is not a shared object */
	LSO_NO_OBJECT,

	/* a file with the name specified exists already */
	LSO_EXISTS,

	/* shared object appears to be damaged */
	LSO_INCONSISTENT,

	/* the operation failed due to an
         * underlying IO error 
         */
	LSO_IO_ERROR,

	/* The object has already been mapped by N processes, so
	 * that there is no PID available for this process.
	 */
	LSO_NO_PID,

	/* there was insufficient memory */
	LSO_NO_MEM,

	/* the name could not be found in the dictionary */
	LSO_NOT_FOUND
} lso_error_t;

#define LSO_USER_DICT 10

/*  lso_error2str(error) - get a symbol for and error code
 */
const char* lso_error2str(lso_error_t error);

#include "so_internal.h"

/* lso_open(file, handle) - open a shared object on disk
 *
 * Open a shared object stored on disk. 
 * In case of failure, the function returns an error code and *handle
 * is set to NULL. In case of success, the function returns LSO_OK and
 * *handle is set to a valid handle.
 */
lso_error_t lso_open(const char* file, lso_t* handle);


/* lso_close(handle) - close a shared object
 *
 * Closes the shared object identified by handle. After closing the
 * object, *handle will be set to an invalid value.
 */
lso_error_t lso_close(lso_t* handle);


/* lso_create(file, number_of_procs, heap_size) - create a new shared object
 * 
 * Create a new shared object on disk if it does not yet exist. The
 * shared object will have a heap size of heap_size bytes and will
 * allow number_of_procs to access the object concurrently.
 *
 * This does not open the object.
 */
lso_error_t lso_create(const char* file, unsigned int number_of_procs, 
		       unsigned int heap_size);


/* lso_malloc(handle, size) - reserve memory on the heap
 *
 * Reserve some memory on the heap. The function returns eiter NULL if
 * there is no more memory or a _global_ pointer to the allocated
 * space. That pointer cannot be used for computations directly as it
 * represents an offset in the shared object.
 * Use lso_local_ptr() to obtain a local pointer.
 */
lso_global_t lso_malloc(lso_t handle, unsigned int size);


/* lso_free(handle, ptr) - free memory on the heap
 *
 * Free a chunk of memory that was allocated earlier with
 * lso_malloc(). This function expects a _global_ pointer.
 */
void lso_free(lso_t handle, lso_global_t ptr);


/* lso_put(handle, name, ptr) - give a pointer a "well-known" name.
 *
 * This facility is provided so that processes can discover "root
 * objects" of the shared data structure.
 *
 * The difference between lso_put() and lso_put_exclusive() is that
 * lso_put() will silently update existing pointers while _exlclusive will fail
 * if the entry already exists.
 */
lso_error_t lso_put(lso_t handle, const char* name, lso_global_t ptr);
lso_error_t lso_put_exclusive(lso_t handle, const char* name, lso_global_t ptr);

/* lso_get(handle, name) - retrieve a well-known ptr.
 *
 * Use it to find root objects.
 */
lso_global_t lso_get(lso_t handle, const char* name);

/* lso_del(handle, name) - delete the name from the dictionary
 */
lso_error_t  lso_del(lso_t handle, const char* name);


/* lso_local_ptr(handle, ptr) - convert a global offset to a local
 * address.
 *
 * Use this to find out where in the address space a pointer in the
 * shared file is mapped.
 */
static inline void*        lso_local_ptr(lso_t handle, lso_global_t ptr)
{
	return ((char*) ((struct litmus_so*) handle)->base) + ptr;
}


/* lso_global_ptr(handle, ptr) - convert a local address to a global
 * offset.
 *
 * Use this to convert local pointers to a globally meaningful format.
 */
static inline lso_global_t lso_global_ptr(lso_t handle, void * ptr)
{
	return ((lso_global_t) (((char*) ptr) - 
				((char*)((struct litmus_so*) handle)->base)));
}


/* lso_get_pid(handle) - determine which pid we got when mapping the 
 *                       the object.
 *
 * PIDs in this context range from 0 to N - 1.
 */
static inline  unsigned int lso_get_pid(lso_t handle)
{
	return ((struct litmus_so*) handle)->pid;
}

#endif
