
#ifndef __SO_QLOCK_H__
#define __SO_QLOCK_H__

#include <sched.h>

#include "so_internal_arch.h"
#include "litmus_so.h"
#include "litmus.h"

#define SMP_CACHE_BYTES 128

struct slot {
	union { 
		/* pad the values being spun on to make sure
		   that they are cache local
		*/
		volatile enum {
			MUST_WAIT,
			HAS_LOCK
		} val;
		char 	padding[SMP_CACHE_BYTES];
	};	
};

struct queuelock {
	lso_global_t	slots;

	/* The next slot that will be available.
	 */
	atomic_t 	next_slot;
};


static inline void __queue_lock_init(struct queuelock *lock, int nr_slots, struct slot* slots)
{
	int i;

	for (i = 0; i < nr_slots; i++)
	        slots[i].val 	= MUST_WAIT;	
	slots[0].val 	= HAS_LOCK;
	atomic_set(&lock->next_slot, 0);
}


static inline void raw_queue_lock(struct litmus_so* lso, struct queuelock *lock, int *spin_slot)
{
 	struct slot* slots =  lso_local_ptr((lso_t) lso, lock->slots);
	volatile struct slot* myslot;

	/* Get slot to spin on. atomic_inc_return() returns the incremented
	 * value, so take one of again
	 */
	*spin_slot = atomic_inc_return(&lock->next_slot) - 1;
	/* check for wrap-around
	 * This could probably optimized away if we ensure that NR_CPUS divides 
	 * INT_MAX...
	 */
	if (*spin_slot == lso->base->max_pid - 1)
		atomic_add(-((int)lso->base->max_pid), &lock->next_slot);
	/* range limit*/
	*spin_slot %= lso->base->max_pid;
	/* spin until you acquire the lock */
	myslot = slots + *spin_slot;

	while (myslot->val == MUST_WAIT)
		__asm__ __volatile__("pause \n\t");

	/* reset the lock */
	slots[*spin_slot].val = MUST_WAIT;
}

static inline void raw_queue_unlock(struct litmus_so* lso, struct queuelock *lock, 
				    int *spin_slot)
{	
	struct slot* slots = lso_local_ptr((lso_t) lso, lock->slots);

	slots[(*spin_slot + 1) % lso->base->max_pid].val = HAS_LOCK;
}

lso_global_t __lso_malloc(struct litmus_so* lso, unsigned int size);
void __lso_free(struct litmus_so* lso, lso_global_t ptr);

static inline lso_global_t __alloc_queuelock(struct litmus_so* lso) 
{
	int nr_slots = lso->base->max_pid;
	lso_global_t slots;
	lso_global_t lock;
	struct queuelock* _lock;
	struct slot*      _slots;

	lock  = __lso_malloc(lso, sizeof(struct queuelock) + 
				       sizeof(struct slot) * nr_slots);
	slots = lock + sizeof(struct queuelock);

	if (!lock)
		return 0;
	
	_lock       = lso_local_ptr((lso_t) lso, lock);
	_slots      = lso_local_ptr((lso_t) lso, slots);
	__queue_lock_init(_lock, nr_slots, _slots);
	_lock->slots = slots;
	return lock;
}

static inline void __free_queuelock(struct litmus_so* lso, lso_global_t lock)
{
	struct queuelock* ql;
	ql = lso_local_ptr((lso_t) lso, lock);
	__lso_free(lso, ql->slots);
	__lso_free(lso, lock);	
}

static inline lso_global_t alloc_queuelock(struct litmus_so* lso) 
{
	lso_global_t ret;
	__lso_lock(lso);
	ret = __alloc_queuelock(lso);
	__lso_unlock(lso);
	return ret;
}

static inline void free_queuelock(struct litmus_so* lso, lso_global_t lock) 
{
	__lso_lock(lso);
	__free_queuelock(lso, lock);
	__lso_unlock(lso);
}


#endif
