
#include <stdlib.h>
#include <string.h>

#include "litmus_so.h"
#include "so_qlock.h"

/* we need access to the raw memory management functions */
lso_global_t __lso_malloc(struct litmus_so* lso, unsigned int size);
void         __lso_free(struct litmus_so* lso, lso_global_t ptr);

#define TABLE_SIZE 149 /* prime */

lso_global_t __allocate_dictionary(struct litmus_so* lso)
{
	lso_global_t* table;
	int i;

	lso->base->dict = __lso_malloc(lso, sizeof(lso_global_t) * TABLE_SIZE);
	table = lso_local_ptr((lso_t) lso, lso->base->dict);
	for (i = 0; i < TABLE_SIZE; i++)
		table[i] = 0;
	return lso->base->dict;
}

/* This is the djb2 hash function as given at:
 *
 *	http://www.cse.yorku.ca/~oz/hash.html
 */
static unsigned long djb2_hash(const char *s_str)
{
	const unsigned char *str = (const unsigned char*) s_str;
	unsigned long hash = 5381;
        int c;
	
        while ((c = *str++))
		hash = ((hash << 5) + hash) + c; /* hash * 33 + c */
	
        return hash;
}

struct chain_cell {
	lso_global_t 	next;
	lso_global_t	val;
};

#define get_key(cell) (((char* ) cell) + sizeof(struct chain_cell))

#define get_next(cell) ( (struct chain_cell*)  (cell->next ? \
	lso_local_ptr((lso_t) lso, cell->next)  : NULL)  )


static lso_global_t alloc_chain_cell(struct litmus_so* lso, const char* key,
				     lso_global_t val, lso_global_t next)
{
	lso_global_t ptr = __lso_malloc(lso, sizeof(struct chain_cell) + strlen(key) + 1);
	struct chain_cell *cell;
	
	if (ptr) {
		cell = lso_local_ptr((lso_t) lso, ptr);
		cell->next = next;
		cell->val  = val;
		strcpy(get_key(cell), key);
	}
	return ptr;
}

static struct chain_cell* search(struct litmus_so* lso, 
				 const char* key, lso_global_t start,
				 struct chain_cell** prev)
{
	struct chain_cell *cell;
	if (prev)
		*prev = NULL;

	if (!start)
		return NULL;
	cell = lso_local_ptr((lso_t) lso, start);
	while (cell && strcmp(get_key(cell), key)) {
		if (prev)
			*prev = cell;
		cell = get_next(cell);
	}
	return cell;
}

static lso_error_t __lso_name(struct litmus_so* lso, const char* key, 
		       lso_global_t val, int excl)
{
	struct chain_cell *cell;
	lso_global_t* table;
	lso_global_t  ncell;
	lso_error_t   ret = LSO_OK;
	int idx = djb2_hash(key) % TABLE_SIZE;

	table = lso_local_ptr((lso_t) lso, lso->base->dict);
	cell  = search(lso, key, table[idx], NULL);
	if (cell) {
		if (!excl) 
		        /* entry exists, just update value */
			cell->val = val;
		else
			/* exclusive mode: don't update */
			ret = LSO_EXISTS;
	}
	else {
		ncell = alloc_chain_cell(lso, key, val, table[idx]);
		if (ncell)
			table[idx] = ncell;		       
		else
			ret = LSO_NO_MEM;
	}
	return ret;
}

static lso_global_t __lso_find(struct litmus_so* lso, const char* key)
{
	struct chain_cell *cell;
	lso_global_t* table;
	int idx = djb2_hash(key) % TABLE_SIZE;
	table = lso_local_ptr((lso_t) lso, lso->base->dict);
	cell  = search(lso, key, table[idx], NULL);
	if (cell)
		return cell->val;
	else
		return 0;
}

static lso_error_t __lso_unname(struct litmus_so* lso, const char* key)
{
	struct chain_cell *cell, *prev;	
	lso_global_t* table;
	int idx = djb2_hash(key) % TABLE_SIZE;
	table = lso_local_ptr((lso_t) lso, lso->base->dict);
	cell  = search(lso, key, table[idx], &prev);
	if (cell) {
		if (prev)
			prev->next = cell->next;
	        else
			table[idx] = cell->next;
		__lso_free(lso, lso_global_ptr((lso_t) lso, cell));
		return LSO_OK;
	}
	else
		return LSO_NOT_FOUND;	
}

lso_error_t lso_put(lso_t handle, const char* key, lso_global_t val)
{
	struct litmus_so* lso = (struct litmus_so*) handle;
	lso_error_t ret;
	__lso_lock(lso);
	ret = __lso_name(lso, key, val, 0);
	__lso_unlock(lso);
	return ret;	
}

lso_error_t lso_put_exclusive(lso_t handle, const char* key, lso_global_t val)
{
	struct litmus_so* lso = (struct litmus_so*) handle;
	lso_error_t ret;
	__lso_lock(lso);
	ret = __lso_name(lso, key, val, 1);
	__lso_unlock(lso);
	return ret;	
}

lso_global_t lso_get(lso_t handle, const char* key)
{
	struct litmus_so* lso = (struct litmus_so*) handle;
	lso_global_t ret;
	__lso_lock(lso);
	ret = __lso_find(lso, key);
	__lso_unlock(lso);
	return ret;	
}

lso_error_t lso_del(lso_t handle, const char* key)
{
	struct litmus_so* lso = (struct litmus_so*) handle;
	lso_error_t ret;
	__lso_lock(lso);
	ret = __lso_unname(lso, key);
	__lso_unlock(lso);
	return ret;	
}
