

#include "litmus_so.h"
#include "so_qlock.h"

/* The memory management for the shared object's heap. This is a straight
 * forward first heap implementation. The heap is protected by the lso->memlock
 * lock.
 *
 * TODO: It would be nice if we support cache aligned allocations.
 */

#define cell(x) ((struct lso_heap_cell*) lso_local_ptr((lso_t) lso, x))
#define invcell(x) ((lso_global_t) (((char*)x) -  ((char*)lso->base)))

/* this function assumes exlusive access to the heap */
/* it does a simple first-fit search for a free heap cell */
lso_global_t __lso_malloc(struct litmus_so* lso, unsigned int size)
{
	struct lso_heap_cell *start = cell(lso->base->heap.free_list_head);
	struct lso_heap_cell *cur, *next;
	unsigned int cut;

	/* null pointer indicates out-of-memory */
	if (!lso->base->heap.free_list_head)
		return 0;

	/* search for a cell with space */
	cur = start;
	do {
		if (cur->size >= size)
			break;
		cur = cell(cur->next);
	} while (cur != start);

	if (cur->size >= size) {
		/* how much left-over space do we have? */
		cut = cur->size - size;
		if (cut > sizeof(struct lso_heap_cell) + 16) {
			/* enough to requeue in free list */
			next       = (struct lso_heap_cell*) 
				((char*) cur + (sizeof(struct lso_heap_cell) + size));
			next->size = cur->size - (sizeof(struct lso_heap_cell) + size);
			if (cell(cur->next) == cur) {
				/* we split a single cell */
				next->prev = invcell(next);
				next->next = invcell(next);
			} else {
				/* there is another cell  */
				next->prev = cur->prev;
				next->next = cur->next;
				cell(cur->next)->prev = invcell(next);
				cell(cur->prev)->next = invcell(next);
			}
			lso->base->heap.free_list_head = invcell(next);
			/* store allocation size */		
			cur->size = size;
		} else {
			if (cell(cur->next) == cur)
				/* free list is now empty */
				lso->base->heap.free_list_head = 0;
			else {
				cell(cur->next)->prev = cur->prev;
				cell(cur->prev)->next = cur->next;				
				lso->base->heap.free_list_head = cur->next;
			}			
		}
		cur->next = 0;
		cur->prev = 0;
		return invcell(cur) + sizeof(struct lso_heap_cell);		
	} else
		return 0;
}


static void merge(struct litmus_so* lso, struct lso_heap_cell* hc)
{
	struct lso_heap_cell *next;
	if (hc->size + invcell(hc) + sizeof(struct lso_heap_cell) == hc->next) {
		if (lso->base->heap.free_list_head == hc->next)
			lso->base->heap.free_list_head = invcell(hc);
		next = cell(hc->next);
		/* merge cells */
		hc->size      += next->size + sizeof(struct lso_heap_cell);
		hc->next       = next->next;
		cell(hc->next)->prev = invcell(hc);
	}
}


void __lso_free(struct litmus_so* lso, lso_global_t ptr)
{
	struct lso_heap_cell *start = cell(lso->base->heap.free_list_head);
	struct lso_heap_cell *cur, *prev, *next;

	cur = cell((ptr - sizeof(struct lso_heap_cell)));
	if (!lso->base->heap.free_list_head) {
		/* there are no other free cells */
		cur->next = invcell(cur);
		cur->prev = invcell(cur);
		lso->base->heap.free_list_head = invcell(cur);		
	} else {
		/* find the right spot for reinsertion */
		if (start == cell(start->next)) {
			/* only one cell */
			cur->next   = invcell(start);
			cur->prev   = invcell(start);
			start->next = invcell(cur);
			start->prev = invcell(cur);
			if (cur < start)
				merge(lso, cur);			
			else
				merge(lso, start);
		} else {
			prev = start;
			next = cell(prev->next);
			do {
				/* scan right */
				while (prev < next && 
				       (next < cur || cur < prev)) {
					prev = next;
					next = cell(next->next);
				}

				if (prev < next)
					/* found it */
					break;

				/* wrap around case */

				if (cur > prev || cur < next)
					/* also found it */
					break;
				/* keep on searching */
				prev = next;
				next = cell(next->next);
			} while (1);
			cur->next = invcell(next);
			cur->prev = invcell(prev);
			next->prev = invcell(cur);
			prev->next = invcell(cur);
			merge(lso, cur);
			merge(lso, prev);			
		}
		
	}
}


#define MIN_SIZE 256

lso_global_t lso_malloc(lso_t handle, unsigned int size)
{
	lso_global_t ret;
	struct litmus_so* lso = (struct litmus_so*) handle;
	if (size < MIN_SIZE)
		size = MIN_SIZE;
	__lso_lock(lso);
	ret = __lso_malloc(lso, size);
	__lso_unlock(lso);
	return ret;
}



void lso_free(lso_t handle, lso_global_t ptr)
{
	struct litmus_so* lso = (struct litmus_so*) handle;
	if (!ptr)
		return;
	__lso_lock(lso);
	__lso_free(lso, ptr);
	__lso_unlock(lso);
}
