/* This is a simple example that shows how to use the LITMUS shared object
 * API. The example implements a uni-directional message queue that can be
 * accessed by multiple producers and multiple consumers.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>

/* These are the headers for libso (|lib|rary for |s|hared |o|bjects). Client
 * applications should not include any of the internal headers directly.
 */
#include "litmus_so.h"      /* shared object basics like open, close, malloc,
			     * free, etc. */
#include "litmus_qlock.h"   /* a userspace queue lock implementation build on
			     * top of the shared object library */

/* a macro that is used in the example to "visualize" the return codes */
#define CALL(s) { err = s; printf("%s\n\t=> %s\n", #s, lso_error2str(err)); }

/* global for the macro above */
lso_error_t err;

/* this is he main handle to the shared object 
 * we keep it global so that we can close the object 
 * properly in case we are interrupted.
 */
lso_t lso;

/* convenience function that properly closes the shared object */
void die(const char* msg) 
{
	fprintf(stderr, "%s\n", msg);
	CALL( lso_close(&lso) );
	exit(1);
}

/* make sure we leave the object in a consistent state if we exit early */
void on_sigint(int sig)
{
	die("Interrupted!");
}

#define SO_NAME 	"msgq.lso"	/* name of the file conaining the
					   shared objec*/
#define LOCK_NAME       "buf.lock"	/* name of the queue lock used to
					   ensure mutual exclusion */
#define BUFFER_NAME	"the buffer"	/* symbolic name used to find the
					   buffer in the heap of the shared object */
#define MAX_PROCS 	4		/* maximum number of processes that can
					   open the object concurrently */
#define MAX_LENGTH 	80
#define MAX_MSG_COUNT	26

/* a very simple fixed size message buffer */
struct simple_message_buffer {	
	char buf[MAX_MSG_COUNT][MAX_LENGTH];

	int read;
	int write;
};

void init_buffer(struct simple_message_buffer* buf)
{
	buf->read  = MAX_MSG_COUNT - 1;
	buf->write = 0;
}

/* assumes exclusive access to the buffer */
int put_msg(struct simple_message_buffer* buf, char* msg)
{
	int next = (buf->write + 1) % MAX_MSG_COUNT;
	if (next != buf->read) {
		strncpy(buf->buf[buf->write], msg, MAX_LENGTH);
		buf->write = next;
		return 1;
	}
	return 0;
}

/* assumes exclusive access to the buffer */
int get_msg(struct simple_message_buffer* buf, char* msg)
{
	int next = (buf->read + 1) % MAX_MSG_COUNT;
	if (next != buf->write) {
		strncpy(msg, buf->buf[next], MAX_LENGTH);
		buf->read = next;
		return 1;
	}
	return 0;
}

void producer(struct simple_message_buffer* buf, lso_qlock_t lock);
void consumer(struct simple_message_buffer* buf, lso_qlock_t lock);

/* The main() function works as follows:
 *
 *  1) open the shared object 
 *     - if it fails try to create it
 *     - this will give us a handle to refer to the object
 *       in later calls.
 *
 *  2) get a handle to the lock
 *     - if we can't find try to create it
 *
 *  3) find the message buffer in the heap
 *     - if we can't find it, try to create it
 *     - we use the dictionary API of libso to find
 *       the buffer and the address translation to 
 *       obtain a local pointer
 *
 *  4) Finally, we start the producer/consumer loop.
 *
 *  5) Don't forget to cleanup on exit, the object state is
 *     persistent!
 */
int main(int argc, char** argv) 
{
	struct simple_message_buffer* buf;
	lso_global_t buf_ptr;
	lso_qlock_t lock;

	/* make sure we don't exit without cleanup */
	signal(SIGINT, on_sigint);

	/* try to obtain a handle */
	CALL( lso_open(SO_NAME, &lso) );

	if (err != LSO_OK) {
		/* that didn't work. Maybe the file is missing */
		printf("Trying to create shared object.\n");
		/* object creation is guaranteed to be race free, even if
		 * several processes try to create the same object in parallel 
		 */
		CALL( lso_create(SO_NAME, MAX_PROCS, 2 * sizeof(struct simple_message_buffer)) );	
		/* now we should be able to open it */
		CALL( lso_open(SO_NAME, &lso) );
	}
	
	if (err == LSO_OK) {
		/* libso assigns every attached process a unique ID in the
		 * range 0 .. N-1 
		 */
		printf("my pid is %u\n", lso_get_pid(lso));

		/* try to find the lock */
		CALL( lso_get_qlock(lso, LOCK_NAME, &lock) );
		if (err != LSO_OK) {
			/* maybe it is missing because the file was just
			 * created */
			printf("trying to create lock\n");
			CALL( lso_create_qlock(lso, LOCK_NAME) );
			CALL( lso_get_qlock(lso, LOCK_NAME, &lock) );		       
			if (err != LSO_OK)
				die("where is the lock?");
		}
		
		/* use the symbol lookup API of libso to find the buffer */
		buf_ptr = lso_get(lso, BUFFER_NAME);
		if (!buf_ptr) {
			/* if it is not there we must create it */
			printf("trying to allocate buffer\n");
			/* use libso memory allocator to reserve space on the
			 *  shared heap */
			buf_ptr = lso_malloc(lso, sizeof(struct simple_message_buffer));
			if (!buf_ptr)
				die("insufficient memory in shared object!");
			/* convert to local (= usable by C-code) pointer */
			buf = lso_local_ptr(lso, buf_ptr);
			init_buffer(buf);
			/* associate symbol with address */
			CALL( lso_put_exclusive(lso, BUFFER_NAME, buf_ptr) );
			if (err != LSO_OK) {
				printf("somebody created the buffer before I did\n");
				lso_free(lso, buf_ptr);
				buf_ptr = lso_get(lso, BUFFER_NAME);
				if (!buf_ptr)
					die("huh? something's broken. Where's my buffer?");
			}
		}
		/* convert to local address */
		buf = lso_local_ptr(lso, buf_ptr);

		/* do work */
		if (argc == 1 || strcmp(argv[1], "producer"))
			consumer(buf, lock);
		else
			producer(buf, lock);

		/* cleanup */
		CALL(lso_close(&lso));
	}

	return 0;
}

/* the producer reads messages and puts them in the buffer */
void producer(struct simple_message_buffer* buf, lso_qlock_t lock)
{
	char msg[MAX_LENGTH];
	int ok;

	printf("This is a producer, reading from STDIN.\n");
	
	/* read a message from stdin */
	while (fgets(msg, MAX_LENGTH, stdin) != NULL) {
		/* make sure we don't get interrupted while holding the lock */
		signal(SIGINT, SIG_IGN);
		/* acquire the lock */
		lso_qlock(lock);
		/* we now have exclusive access to the buffer */
		/* try insert a message */
		ok = put_msg(buf, msg);
		/* release lock */
		lso_qunlock(lock);
		/* reenable "interrupts" */
		signal(SIGINT, on_sigint);
		/* did we find sufficient space in the buffer? */
		if (ok)
			printf("\tOK\n");
		else
			printf("\tFAILED\n");
	}
}

/* the consumer tries once a seconde to receive messages from the buffer */
void consumer(struct simple_message_buffer* buf, lso_qlock_t lock)
{
	char msg[MAX_LENGTH];
	int got_one;
	
	printf("This is a consumer, writing to STDOUT.\n");

	do {		
		do {
			/* disable interrupts */
			signal(SIGINT, SIG_IGN);
			/* enter critical section */
			lso_qlock(lock);
			/* look for message */
			got_one = get_msg(buf, msg);
			/* exit critical section */
			lso_qunlock(lock);
			/* reenable interupts */
			signal(SIGINT, on_sigint);
			/* output received message */
			if (got_one)
				printf("%s", msg);
		} while (got_one);
		/* don't busy wait for new messages */
		sleep(1);
	} while (1);
}
